@@ -31,12 +31,12 @@
 
 #include "config.h"
 
-#include "galias.h"
 #include "glibconfig.h"
 
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <wchar.h>
 #include <errno.h>
 
 #define STRICT			/* Strict typing, please */
@@ -52,6 +52,7 @@
 #endif /* _MSC_VER */
 
 #include "glib.h"
+#include "galias.h"
 
 #ifdef G_WITH_CYGWIN
 #include <sys/cygwin.h>
@@ -1064,7 +1065,7 @@ g_win32_error_message (gint error)
 		      |FORMAT_MESSAGE_IGNORE_INSERTS
 		      |FORMAT_MESSAGE_FROM_SYSTEM,
 		      NULL, error, 0,
-		      (LPTSTR) &msg, 0, NULL);
+		      (LPWSTR) &msg, 0, NULL);
       if (msg != NULL)
 	{
 	  nchars = wcslen (msg);
@@ -1198,7 +1199,7 @@ get_package_directory_from_module (gchar *module_name)
 /**
  * g_win32_get_package_installation_directory:
  * @package: An identifier for a software package, or %NULL, in UTF-8
- * @dll_name: The name of a DLL that a package provides, or %NULL, in the GLib file name encoding, which is UTF-8 on Windows.
+ * @dll_name: The name of a DLL that a package provides, or %NULL, in UTF-8
  *
  * Try to determine the installation directory for a software package.
  * Typically used by GNU software packages.
@@ -1222,7 +1223,7 @@ get_package_directory_from_module (gchar *module_name)
  * @dll_name was %NULL.
  *
  * If both @package and @dll_name are %NULL, the directory from where
- * the main executable of the process was loaded is uses instead in
+ * the main executable of the process was loaded is used instead in
  * the same way as above.
  *
  * Returns: a string containing the installation directory for
@@ -1317,7 +1318,7 @@ g_win32_get_package_installation_directory (gchar *package,
 	{
 	  g_hash_table_insert (package_dirs, package, result);
 	  G_UNLOCK (package_dirs);
-	  return result;
+	  return g_strdup (result);
 	}
       G_UNLOCK (package_dirs);
     }
@@ -1331,11 +1332,41 @@ g_win32_get_package_installation_directory (gchar *package,
   return result;
 }
 
+#undef g_win32_get_package_installation_directory
+
+/* DLL ABI binary compatibility version that uses system codepage file names */
+
+gchar *
+g_win32_get_package_installation_directory (gchar *package,
+					    gchar *dll_name)
+{
+  gchar *utf8_package = NULL, *utf8_dll_name = NULL;
+  gchar *utf8_retval, *retval;
+
+  if (package != NULL)
+    utf8_package = g_locale_to_utf8 (package, -1, NULL, NULL, NULL);
+
+  if (dll_name != NULL)
+    utf8_dll_name = g_locale_to_utf8 (dll_name, -1, NULL, NULL, NULL);
+
+  utf8_retval =
+    g_win32_get_package_installation_directory_utf8 (utf8_package,
+						     utf8_dll_name);
+
+  retval = g_locale_from_utf8 (utf8_retval, -1, NULL, NULL, NULL);
+
+  g_free (utf8_package);
+  g_free (utf8_dll_name);
+  g_free (utf8_retval);
+
+  return retval;
+}
+
 /**
  * g_win32_get_package_installation_subdirectory:
- * @package: An identifier for a software package, or %NULL.
- * @dll_name: The name of a DLL that a package provides, or %NULL.
- * @subdir: A subdirectory of the package installation directory.
+ * @package: An identifier for a software package, in UTF-8, or %NULL
+ * @dll_name: The name of a DLL that a package provides, in UTF-8, or %NULL
+ * @subdir: A subdirectory of the package installation directory, also in UTF-8
  *
  * Returns a newly-allocated string containing the path of the
  * subdirectory @subdir in the return value from calling
@@ -1343,11 +1374,31 @@ g_win32_get_package_installation_directory (gchar *package,
  * @dll_name parameters. 
  *
  * Returns: a string containing the complete path to @subdir inside
- * the installation directory of @package. The string is in the GLib
- * file name encoding, i.e. UTF-8 on Windows. The return value should
- * be freed with g_free() when no longer needed.
+ * the installation directory of @package. The returned string is in
+ * the GLib file name encoding, i.e. UTF-8 on Windows. The return
+ * value should be freed with g_free() when no longer needed.
  **/
 
+gchar *
+g_win32_get_package_installation_subdirectory (gchar *package,
+					       gchar *dll_name,
+					       gchar *subdir)
+{
+  gchar *prefix;
+  gchar *dirname;
+
+  prefix = g_win32_get_package_installation_directory_utf8 (package, dll_name);
+
+  dirname = g_build_filename (prefix, subdir, NULL);
+  g_free (prefix);
+
+  return dirname;
+}
+
+#undef g_win32_get_package_installation_subdirectory
+
+/* DLL ABI binary compatibility version that uses system codepage file names */
+
 gchar *
 g_win32_get_package_installation_subdirectory (gchar *package,
 					       gchar *dll_name,
@@ -1415,3 +1466,61 @@ g_win32_get_windows_version (void)
   
   return windows_version;
 }
+
+/**
+ * g_win32_locale_filename_from_utf8:
+ * @utf8filename: a UTF-8 encoded filename.
+ *
+ * Converts a filename from UTF-8 to the system codepage.
+ *
+ * On NT-based Windows, on NTFS file systems, file names are in
+ * Unicode. It is quite possible that Unicode file names contain
+ * characters not representable in the system codepage. (For instance,
+ * Greek or Cyrillic characters on Western European or US Windows
+ * installations, or various less common CJK characters on CJK Windows
+ * installations.)
+ *
+ * In such a case, and if the filename refers to an existing file, and
+ * the file system stores alternate short (8.3) names for directory
+ * entries, the short form of the filename is returned. Note that the
+ * "short" name might in fact be longer than the Unicode name if the
+ * Unicode name has very short pathname components containing
+ * non-ASCII characters. If no system codepage name for the file is
+ * possible, %NULL is returned.
+ *
+ * The return value is dynamically allocated and should be freed with
+ * g_free() when no longer needed.
+ *
+ * Return value: The converted filename, or %NULL on conversion
+ * failure and lack of short names.
+ *
+ * Since: 2.8
+ */
+gchar *
+g_win32_locale_filename_from_utf8 (const gchar *utf8filename)
+{
+  gchar *retval = g_locale_from_utf8 (utf8filename, -1, NULL, NULL, NULL);
+
+  if (retval == NULL && G_WIN32_HAVE_WIDECHAR_API ())
+    {
+      /* Conversion failed, so convert to wide chars, check if there
+       * is a 8.3 version, and use that.
+       */
+      wchar_t *wname = g_utf8_to_utf16 (utf8filename, -1, NULL, NULL, NULL);
+      if (wname != NULL)
+	{
+	  wchar_t wshortname[MAX_PATH + 1];
+	  if (GetShortPathNameW (wname, wshortname, G_N_ELEMENTS (wshortname)))
+	    {
+	      gchar *tem = g_utf16_to_utf8 (wshortname, -1, NULL, NULL, NULL);
+	      retval = g_locale_from_utf8 (tem, -1, NULL, NULL, NULL);
+	      g_free (tem);
+	    }
+	  g_free (wname);
+	}
+    }
+  return retval;
+}
+
+#define __G_WIN32_C__
+#include "galiasdef.c"
